// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "il_writer_utils.h"
#include "memory.h"

/**
 \addtogroup atomic atomic operations
atomic functions are used to atomically modify values in global or local memory.

 example:
 \code{.cpp}
 struct foo :
 kernel<foo>
 {
    void program(resource<int>& a, resource<pair<uint64_t, double>>& B)
    {
        atomic_add(a[0], 1);                               // increments a[0] by 1
        gpu_double x = atomic_xchg(b[12].second, 3.14);    // Writes 3.14 at B[12].second, returns in x the previous value
    }
 };
 \endcode

 \{
*/

namespace goopax {
namespace impl {
namespace interface {
extern "C" {
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_atomic_func(gpu_writer_base_impl*, gpu_writer_base_impl*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) errno_type goopax_41E_atomic_func(gpu_writer_base_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_atomic_cmpxchg(gpu_writer_base_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
}
}


template<class gpu_A, class gpu_B>
gpu_A call_atomic(gpu_A& ref, const gpu_B& value, unsigned int N)
{
    //static_assert(!is_pointer<typename gettype<gpu_B>::cpu>::value);
    gpu_A ret;
    check_errno(interface::goopax_41E_atomic_func(ref.impl.get(), value.impl.get(), ret.impl.get(), N));
    return ret;
}


/// atomic add
/// \ingroup atomic
/// Depending on the hardware, floating point operations may be emulated and will be slower.
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to add
/// \return previous value
template<typename T>
typename enable_if<gettype<T>::size == 4 || gettype<T>::size == 8, gpu_type<T>>::type
atomic_add(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 0);
}
template<typename T>
gpu_type<T*> atomic_add(gpu_type<T*>& ref, const gpu_size_t value)
{
    return call_atomic<gpu_type<T*>>(ref, value * sizeof(T), 0);
}


/// atomic min
/// \ingroup atomic
/// Depending on the hardware, floating point operations may be emulated and will be slower.
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to perform min operation with
/// \return previous value
template<typename T>
typename enable_if<gettype<T>::size == 4 || gettype<T>::size == 8, gpu_type<T>>::type
atomic_min(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 4);
}


/// atomic max
/// \ingroup atomic
/// Depending on the hardware, floating point operations may be emulated and will be slower.
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to perform max operation with
/// \return previous value
template<typename T>
typename enable_if<gettype<T>::size == 4 || gettype<T>::size == 8, gpu_type<T>>::type
atomic_max(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 5);
}

/// atomic and
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to perform and operation with
/// \return previous value
template<typename T>
typename enable_if<(gettype<T>::size == 4 || gettype<T>::size == 8) && is_integral<T>::value, gpu_type<T>>::type
atomic_and(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 1);
}

/// atomic or
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to perform or operation with
/// \return previous value
template<typename T>
typename enable_if<(gettype<T>::size == 4 || gettype<T>::size == 8) && is_integral<T>::value, gpu_type<T>>::type
atomic_or(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 2);
}

/// atomic xor
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to perform xor operation with
/// \return previous value
template<typename T>
typename enable_if<(gettype<T>::size == 4 || gettype<T>::size == 8) && is_integral<T>::value, gpu_type<T>>::type
atomic_xor(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 3);
}


/// atomic xchg
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param value value to write
/// \return previous value
template<typename T>
gpu_type<T> atomic_xchg(gpu_type<T>&                                        ref,
                        typename conditional<true, gpu_type<T>, void>::type value)
{
    return call_atomic<gpu_type<T>>(ref, value, 6);
}

/// atomic compare and exchange
/// \ingroup atomic
/// If ref == cmp, write ref = value. Otherwise do nothing.
/// \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \param cmp value to compare with
/// \param value value to write
/// \return previous value
template<typename T>
gpu_type<T> atomic_cmpxchg(gpu_type<T>&                                        ref,
                           typename conditional<true, gpu_type<T>, void>::type cmp,
                           typename conditional<true, gpu_type<T>, void>::type value)
{
    gpu_type<T> ret = value;
    check_errno(interface::goopax_4E_atomic_cmpxchg(ref.impl.get(), cmp.impl.get(), ret.impl.get()));
    return ret;
}


/// atomic load
/// \ingroup atomic
///  \param ref reference to local or global memory address. The value referenced can be of 32 or 64 bit integral type, floating point type, or pointer.
/// \return value at location 'ref'
template<typename T>
typename enable_if<gettype<T>::size == 4 || gettype<T>::size == 8, gpu_type<T>>::type
atomic_load(gpu_type<T>& ref)
{
    return atomic_add(ref, 0);
}

inline gpu_uint32 atomic_load(const gpu_uint32& ref)
{
    return atomic_add(const_cast<gpu_uint32&>(ref), 0);
}

template<typename T>
typename enable_if<gettype<T>::size == 4 || gettype<T>::size == 8, gpu_type<T>>::type
atomic_sub(gpu_type<T>& ref, typename conditional<true, gpu_type<T>, void>::type value)
{
    return atomic_add(ref, -value);
}


}
}


/// \}
