// -*- C++ -*-

#pragma once

#include "pointer.h"
#include "struct.h"

#if GOOPAX_HAVE_METAL && defined(__OBJC__)
@protocol MTLBuffer;
#endif

/**
 \addtogroup memory memory resources
 */

namespace goopax {
namespace impl {

template<typename S, typename T, typename size_type>
[[deprecated("This function will be removed.")]]
typename enable_if<is_same<S, std::ostream>::value, S>::type&
operator<<(S& s, const buffer<T, size_type>& b)
{
    /// Output the buffer content.
    auto               bm        = b.to_vector();
    const unsigned int blocksize = 256;
    for (size_t k = 0; k < b.size(); ++k)
    {
        if (k % blocksize != 0)
            s << " ";
        else if (k != 0)
            s << "\n";
        s << bm[k];
    }
    s << "\n";
    return s;
}

/// \ingroup memory
/// Resource type, used to access buffers from device code. Resources are
/// usually provided as arguments for gpu kernels. The template parameters T and
/// SIZE_TYPE must match the corresponding template parameters of the buffer.
/// \tparam T value type. This can either be a CPU intrinsic type, or a goopax
/// struct type. \tparam SIZE_TYPE This can be a 32 bit or 64 bit integer type.
/// \tparam is_const only read access is allowed. This has the same effect as
/// using a const reference of the resource as kernel parameter.
template<typename T, typename SIZE_TYPE = unsigned int, bool is_const = false>
class resource
    : private gpu_struct::resource_type_switch<
          T, typename gettype<SIZE_TYPE>::cpu, memory::none>::uav_type
{
    static_assert(
        gettype<SIZE_TYPE>::size == 4 || gettype<SIZE_TYPE>::size == 8,
        "Only 32 and 64 bit unsigned integers are allowed as size_type.");
    static_assert(
        is_unsigned<SIZE_TYPE>::value,
        "Only 32 and 64 bit unsigned integers are allowed as size_type.");
    static_assert(
        !is_same<typename unrangetype_core<T>::type, bool>::value,
        "Sorry, buffer<bool> not allowed. Use some other storage type.");
    static_assert(!std::is_const<T>::value,
                  "Please use non-const, non-volatile types.");
    static_assert(!std::is_volatile<T>::value,
                  "Please use non-const, non-volatile types.");

    public:
    using base_t = typename gpu_struct::resource_type_switch<
        T, typename gettype<SIZE_TYPE>::cpu, memory::none>::uav_type;
    using buffer_type = buffer<T, SIZE_TYPE>;

    using typename base_t::const_reference;
    using typename base_t::reference;
    using typename base_t::size_type;
    using typename base_t::ssize_type;
    using typename base_t::value_type;

    using iterator       = gpu_type<T*>;
    using const_iterator = gpu_type<const T*>;

    /// access to element
    using base_t::operator[];
    using base_t::barrier;
    using base_t::size;

    resource_impl* get_impl() { return base_t::base.impl.get(); }

    /// \name iterator access
    /// \{

    iterator begin()
    {
        /// returns begin() pointer
        iterator ret;
        check_errno(
            interface::goopax_4E_get_ptr(this->base.impl.get(), ret.impl.get()));
        return ret;
    }
    const_iterator begin() const
    {
        /// returns begin() pointer
        const_iterator ret;
        check_errno(
            interface::goopax_4E_get_ptr(this->base.impl.get(), ret.impl.get()));
        return ret;
    }

    iterator end()
    {
        /// pointer to the end of the resource
        return begin() + this->size();
    }
    const_iterator end() const
    {
        /// pointer to the end of the resource
        return begin() + this->size();
    }

    const_iterator cbegin() const
    {
        return begin();
    }
    const_iterator cend() const
    {
        return end();
    }
    /// \}

    reference front() { return (*this)[0]; }

    const_reference front() const { return (*this)[0]; }

    reference back() { return (*this)[this->size() - 1]; }

    const_reference back() const { return (*this)[this->size() - 1]; }

    template<memory::scope BSCOPE, typename EIF = void,
             typename = typename enable_if<!is_const, EIF>::type>
    void copy(const local_scope_mem<T, BSCOPE>&              src,
              typename local_scope_mem<T, BSCOPE>::size_type count,
              typename local_scope_mem<T, BSCOPE>::size_type start_dest,
              size_type                                      start_src = 0)
    {
        /// copy data from local memory or private memory to the resource (i.e. to
        /// global memory) For local memory, the copy operation is parallelized over
        /// the workgroup. For private memory, the copy operation is performed for
        /// the individual threads. \param src source \param count number of values
        /// to copy \param start_dest offset in resource \param start_src offset in
        /// source memory
        if (BSCOPE == memory::threadgroup)
        {
            gpu_for_local(0, count, [&](size_type k) {
                (*this)[start_dest + k] = src[start_src + k];
            });
        }
        else if (BSCOPE == memory::thread)
        {
            gpu_for(0, count, [&](size_type k) {
                (*this)[start_dest + k] = src[start_src + k];
            });
        }
    }

    resource(const resource&) = delete;
    resource& operator=(const resource&) = delete;
    /// move semantics
    resource(resource&&) = default;
    /// move semantics
    resource& operator=(resource&&) = default;

    /// \name constructors
    /// Resources normally need not be constructed by the user. Instead, they are
    /// passed as references to the gpu kernel. Resources can, however, be
    /// instantiated to be linked to a specific buffer
    /// \{

    resource()
    {
        /// default constructor. The resulting resource cannot be used. It has to be
        /// replaced by a resource that is linked to a buffer before it can be used.
    }

    resource(buffer_type& a)
        : base_t(a.get_buf())
    {
        /// constructor. The resource is linked to the provided buffer
        /// \param a buffer to link resource to. The buffer must exist and be
        /// properly initialized before the kernel is first called.
        assert(a.get_buf() != nullptr);
    }

    template<typename EIF = void,
             typename     = typename enable_if<is_const, EIF>::type>
    resource(const buffer_type& a)
        : base_t(a.get_buf())
    {
        /// constructor. The resource is linked to the provided const buffer
        /// This constructor is only available if is_const=true.
        /// \param a buffer to link resource to. The buffer must exist and be
        /// properly initialized before the kernel is first called.
        assert(a.get_buf() != nullptr);
    }
};

template<typename T, typename SIZE_TYPE = unsigned int>
using const_resource = resource<T, SIZE_TYPE, true>;

#if __cplusplus >= 201700
template<typename T, typename size_type>
resource(const buffer<T, size_type>&) -> resource<T, size_type, true>;
#endif

template<typename T, typename BASE>
struct debug_data
{
    vector<T> debugvec;
    BASE*     base_begin;

    debug_data(BASE* begin, BASE* end)
        : base_begin(begin)
    {
        debugvec = reinterpret<vector<T>>(vector<BASE>(begin, end));
    }
    ~debug_data()
    {
        vector<BASE> store = reinterpret<vector<BASE>>(debugvec);
        std::copy(store.begin(), store.end(), base_begin);
    }
    debug_data(const debug_data&) = delete;
    debug_data& operator=(const debug_data&) = delete;
};

template<typename T, typename BASE, bool is_checknan>
struct buffer_debug_data;

template<typename T, typename BASE>
struct buffer_debug_data<T, BASE, false>
{};

template<typename T, typename BASE>
struct buffer_debug_data<T, BASE, true>
{
    mutable mutex                         M;
    mutable weak_ptr<debug_data<T, BASE>> data;

    buffer_debug_data() = default;

    buffer_debug_data(const buffer_debug_data& b)
        : data(b.data)
    {}

    buffer_debug_data& operator=(const buffer_debug_data& b)
    {
        data = b.data;
        return *this;
    }
};

template<typename T, typename BASE, bool is_checknan>
struct buffer_map_debug_data;

template<typename T, typename BASE>
struct buffer_map_debug_data<T, BASE, false>
{};

template<typename T, typename BASE>
struct buffer_map_debug_data<T, BASE, true>
{
    shared_ptr<debug_data<T, BASE>> data;
};

/// \ingroup memory
/// buffer map. Can be used to access buffers from host code.
/// For performance reasons, buffer maps should not be called within a loop, but
/// before the loop, if possible. buffer maps must be out of scope before the
/// buffer is used in a kernel call or in another operation. \tparam T value
/// type, must match the buffer value type \tparam is_const if true, allows
/// mapping of const buffers
template<typename T, BUFFER_FLAGS flags>
class buffer_map_base : private map_base
{
    public:
    using value_type               = T;
    using iterator                 = T*;
    using const_iterator           = const T*;
    static constexpr bool is_const = ((flags & BUFFER_WRITE) == 0);

    private:
    value_type* begin_i   = nullptr;
    value_type* end_i     = nullptr;
    using base_value_type = typename buffer<T>::base_value_type;
    buffer_map_debug_data<T, base_value_type,
                          !is_same<T, typename unrangetype<T>::type>::value>
        debugdata;

    template<typename BUF>
    inline void lock_map(BUF& buf, false_type, size_t begin, size_t end)
    {
        void*   b;
        texbuf* buf_use = (begin != end ? buf.get_buf() : nullptr);
        check_errno(
            interface::goopax_43E_tex_lock_map(buf_use, &b, begin, end, flags));
        begin_i   = static_cast<value_type*>(b);
        end_i     = begin_i + (end - begin) / gettype<value_type>::size;
        this->buf = buf_use;
    }
    template<typename BUF>
    void lock_map(BUF& buf, true_type, size_t begin, size_t end)
    {
        void*   b;
        texbuf* buf_use = (begin != end ? buf.get_buf() : nullptr);
        check_errno(
            interface::goopax_43E_tex_lock_map(buf_use, &b, begin, end, flags));
        {
            lock_guard<mutex> LOCK(buf.debugdata.M);
            if (buf.debugdata.data.expired())
            {
                auto* begin_base = static_cast<base_value_type*>(b);
                static_assert(
                    gettype<value_type>::size % gettype<base_value_type>::size == 0,
                    "");
                this->debugdata.data = std::make_shared<debug_data<T, base_value_type>>(
                    begin_base,
                    begin_base + (end - begin) / gettype<base_value_type>::size);
                buf.debugdata.data = this->debugdata.data;
            }
            else
            {
                this->debugdata.data = buf.debugdata.data.lock();
            }
            begin_i = debugdata.data->debugvec.data();
        }
        end_i     = begin_i + (end - begin) / gettype<value_type>::size;
        this->buf = buf_use;
    }

    public:
    template<typename EIF = void,
             typename     = typename enable_if<!is_const, EIF>::type>
    inline value_type* begin()
    {
        /// begin pointer, const
        return begin_i;
    }
    template<typename EIF = void,
             typename     = typename enable_if<!is_const, EIF>::type>
    inline value_type* end()
    {
        /// end pointer, const
        return end_i;
    }
    inline const value_type* begin() const
    {
        /// begin pointer
        return begin_i;
    }
    inline const value_type* end() const
    {
        /// end pointer
        return end_i;
    }

    inline const_iterator cbegin() const
    {
        return begin();
    }
    inline const_iterator cend() const
    {
        return end();
    }

    inline size_t size() const
    {
        /// size of memory region
        return end() - begin();
    }
    inline bool empty() const
    {
        return (end() == begin());
    }

    template<typename EIF = void,
             typename     = typename enable_if<!is_const, EIF>::type>
    inline value_type* data()
    {
        /// same as begin()
        return begin_i;
    }
    inline const value_type* data() const
    {
        /// same as begin()
        return begin_i;
    }

    template<typename EIF = void,
             typename     = typename enable_if<!is_const, EIF>::type>
    inline value_type& operator[](size_t k)
    {
        /// accesses element k
        /// \param k element to get reference to
        return begin()[k];
    }
    inline const value_type& operator[](size_t k) const
    {
        /// accesses element k
        /// \param k element to get const reference to
        return begin()[k];
    }

    protected:
    template<typename SIZE_TYPE>
    buffer_map_base(const buffer<T, SIZE_TYPE>& buf, size_t begin, size_t end)
    {
        assert(begin <= end);
        assert(end <= buf.size());
        lock_map(
            buf,
            integral_constant<bool,
                              !is_same<T, typename unrangetype<T>::type>::value>(),
            begin * gettype<T>::size, end * gettype<T>::size);
    }

    template<typename SIZE_TYPE,
             typename = typename enable_if<is_const, SIZE_TYPE>::type>
    buffer_map_base(const buffer<T, SIZE_TYPE>& buf)
        : buffer_map_base(buf, 0, buf.size())
    {
    }


    buffer_map_base()
    {
    }
};

template<typename T>
class buffer_map : public buffer_map_base<T, BUFFER_READ_WRITE>
{
    public:
    template<typename SIZE_TYPE>
    buffer_map(buffer<T, SIZE_TYPE>& buf, size_t begin, size_t end)
        : buffer_map_base<T, BUFFER_READ_WRITE>(buf, begin, end)
    {}

    template<typename SIZE_TYPE>
    buffer_map(buffer<T, SIZE_TYPE>& buf)
        : buffer_map(buf, 0, buf.size())
    {
        /// constructor. Maps the buffer to host memory.
        /// \param buf buffer to map.
    }

    buffer_map() = default;
};

template<typename T>
class const_buffer_map : public buffer_map_base<T, BUFFER_READ>
{
    public:
    template<typename SIZE_TYPE>
    const_buffer_map(const buffer<T, SIZE_TYPE>& buf, size_t begin, size_t end)
        : buffer_map_base<T, BUFFER_READ>(buf, begin, end)
    {}

    template<typename SIZE_TYPE>
    const_buffer_map(const buffer<T, SIZE_TYPE>& buf)
        : const_buffer_map(buf, 0, buf.size())
    {
        /// constructor. Maps the buffer to host memory.
        /// \param buf buffer to map.
    }

    const_buffer_map() = default;
};


/// \ingroup memory
/// buffer class. Used to allocate memory on the device
/// \tparam T value type. This can either be a CPU intrinsic type, or a goopax
/// struct type. \tparam SIZE_TYPE This can be a 32 bit or 64 bit integer type.
/// 32 bit types give better performance. 64 bit types must be used for very
/// large memory buffers. Goopax will throw an exception if buffers are
/// allocated that are too large for the given size type. The default is 32 bit
/// unsigned int. \tparam is_const only read access is allowed. This has the
/// same effect as using a const reference of the resource as kernel parameter.
template<typename T, typename SIZE_TYPE>
class buffer
{
    static_assert(
        gettype<SIZE_TYPE>::size == 4 || gettype<SIZE_TYPE>::size == 8,
        "Only 32 and 64 bit unsigned integers are allowed as size_type.");
    static_assert(
        is_unsigned<SIZE_TYPE>::value,
        "Only 32 and 64 bit unsigned integers are allowed as size_type.");
    static_assert(!std::is_const<T>::value,
                  "Please use non-const, non-volatile types.");
    static_assert(!std::is_volatile<T>::value,
                  "Please use non-const, non-volatile types.");

    template<typename U, BUFFER_FLAGS flags>
    friend class buffer_map_base;
    friend struct texbuf;
    template<typename P, memory::scope scope>
    friend class gpu_type;

    template<typename TO, typename TFROM, typename SIZE_FROM>
    friend inline TO reinterpret(const buffer<TFROM, SIZE_FROM>& from);

    using base_value_type =
        typename gpu_struct::resource_type_switch<T, SIZE_TYPE,
                                                  memory::none>::d_basetype;

    using base_t = texture<base_value_type, SIZE_TYPE>;

    base_t base;

    public:
    using value_type = T;
    using size_type  = typename base_t::size_type;
    using ssize_type = typename base_t::ssize_type;

    inline texbuf*       get_buf() const { return base.buf; }
    inline goopax_device get_device() const
    {
        /// Returns device on which the buffer is allocated.
        return base.get_device();
    }

    inline size_type size() const
    {
        /// Buffer size
        return base.bytesize() / gettype<T>::size;
    }

    public:
    /// type of resource that matches to the buffer type
    using resource_type = resource<T, SIZE_TYPE>;

    private:
    buffer_debug_data<T, base_value_type,
                      !is_same<T, typename unrangetype<T>::type>::value>
        debugdata;

    public:
    [[deprecated]] inline array<unsigned int, 8> get_hash() const
    {
        return base.get_hash();
    }
    [[deprecated]] inline array<unsigned int, 8> get_hash_ignore_uninit() const
    {
        return base.get_hash_ignore_uninit();
    }

    using map_type       = buffer_map<T>;
    using const_map_type = const_buffer_map<T>;

    bool empty() const { return (this->size() == 0); }

    template<typename SRC_SIZE_TYPE>
    goopax_future<void>
    copy(const buffer<T, SRC_SIZE_TYPE>&               src,
         typename enable_if<true, SRC_SIZE_TYPE>::type count,
         SIZE_TYPE                                     start_dest,
         typename enable_if<true, SRC_SIZE_TYPE>::type start_src)
    {
        /// copy data from one buffer to another buffer
        /// \param src source buffer
        /// \param count number of elements to copy
        /// \param start_dest start position in destination buffer
        /// \param start_src start position in source buffer

        goopax_future<void> ret;
        if constexpr (!is_same<T, typename unrangetype<T>::type>::value)
        {
            vector<T> tmp(count);
            src.copy_to_host(tmp.data(), start_src, start_src + count);
            this->copy_from_host(tmp.data(), start_dest, start_dest + count);
            ret.did_finish = true;
        }
        else
        {
            rundata* rd =
                if_call(interface::goopax_41E_copy, src.get_buf(), this->get_buf(),
                        count * gettype<T>::size, start_dest * gettype<T>::size,
                        start_src * gettype<T>::size);
            ret.set(*rd);
        }
        return ret;
    }
    template<typename SRC_SIZE_TYPE>
    goopax_future<void> copy(const buffer<T, SRC_SIZE_TYPE>& src)
    {
        /// copy data from one buffer to another buffer
        /// \param src source buffer
        if (src.size() != this->size())
        {
            exwhat() << "Memory copy between buffers of different size is not "
                        "allowed without specifying the size!";
            throw EX::syntax();
        }
        return copy(src, src.size(), 0, 0);
    }

    goopax_future<void> copy_to_host_async(T* p, size_t beginpos,
                                           size_t endpos) const
    {
        /// copy data to host asynchronously
        /// The memory region specified can only be used after completion of the
        /// operation. \param p pointer in host memory \param beginpos start
        /// position in device memory \param endpos next to last position in device
        /// memory \return event that can be used to wait for completion.

        goopax_future<void> ret;
        if constexpr (!is_same<T, typename unrangetype<T>::type>::value)
        {
            const_buffer_map<T> map(*this, beginpos, endpos);
            std::copy(map.begin(), map.end(), p);
            ret.did_finish = true;
        }
        else
        {
            rundata* rd = if_call(interface::goopax_41E_copy_to_host, this->get_buf(),
                                  p, beginpos * gettype<T>::size,
                                  (endpos - beginpos) * gettype<T>::size);
            ret.set(*rd);
        }
        return ret;
    }
    inline void copy_to_host(T* p, size_t beginpos, size_t endpos) const
    {
        /// copy data to host
        /// \param p pointer in host memory
        /// \param beginpos start position in device memory
        /// \param endpos next to last position in device memory

        copy_to_host_async(p, beginpos, endpos).wait();
    }

    goopax_future<void> copy_to_host_async(T* p) const
    {
        /// copy data to host asynchronously
        /// The memory region specified can only be used after completion of the
        /// operation. \param p pointer in host memory \return event that can be
        /// used to wait for completion.
        return copy_to_host_async(p, 0, this->size());
    }
    void copy_to_host(T* p) const
    {
        /// copy data to host
        /// \param p pointer in host memory
        copy_to_host_async(p).wait();
    }

    goopax_future<void> copy_from_host_async(const T* p, size_t beginpos,
                                             size_t endpos)
    {
        /// copy data from host memory asynchronously
        /// The memory region specified can only be used after completion of the
        /// operation. \param p pointer in host memory \param beginpos start
        /// position in device memory \param endpos next to last position in device
        /// memory \return event that can be used to wait for completion.
        goopax_future<void> ret;
        if constexpr (!is_same<T, typename unrangetype<T>::type>::value)
        {
            buffer_map<T> map(*this, beginpos, endpos);
            std::copy(p, p + (endpos - beginpos), map.begin());
            ret.did_finish = true;
        }
        else
        {
            rundata* rd = if_call(interface::goopax_41E_copy_from_host,
                                  this->get_buf(), p, beginpos * gettype<T>::size,
                                  (endpos - beginpos) * gettype<T>::size);
            ret.set(*rd);
        }
        return ret;
    }
    void copy_from_host(const T* p, size_t beginpos, size_t endpos)
    {
        /// copy data from host memory
        /// \param p pointer in host memory
        /// \param beginpos start position in device memory
        /// \param endpos next to last position in device memory
        copy_from_host_async(p, beginpos, endpos).wait();
    }
    goopax_future<void> copy_from_host_async(const T* p)
    {
        /// copy data from host asynchronously
        /// The memory region specified can only be used after completion of the
        /// operation. \param p pointer in host memory \return event that can be
        /// used to wait for completion.
        return copy_from_host_async(p, 0, this->size());
    }
    void copy_from_host(const T* p)
    {
        /// copy data from host
        /// \param p pointer in host memory
        copy_from_host_async(p).wait();
    }

    template<class V, class ENABLE = typename enable_if<std::is_base_of<
                          std::random_access_iterator_tag,
                          typename std::iterator_traits<typename V::iterator>::
                              iterator_category>::value>::type>
    [[deprecated("copy assignment from lvalue vector will be removed. Consider passing as rvalue.")]] buffer&
    operator=(const V& v)
    {
        return *this = V(v);
    }

    template<class V, class ENABLE = typename enable_if<std::is_base_of<
                          std::random_access_iterator_tag,
                          typename std::iterator_traits<typename V::iterator>::
                              iterator_category>::value>::type>
    buffer& operator=(V&& v)
    {
        /// copy data from std::vector
        if (v.size() != this->size())
        {
            exwhat() << "Assignment to buffer from container of different size is "
                        "not allowed!";
            throw EX::syntax();
        }
        auto lock = std::make_shared<V>(move(v));
        copy_from_host_async(lock->data()).set_callback([lock] {});
        return *this;
    }

    goopax_future<void> fill(T value, size_type begin, size_type end)
    {
        /// Fill buffer with value
        /// \param value value to fill
        /// \param begin begin of fill region
        /// \param end next to last element to fill
        goopax_future<void> ret;
        if constexpr (!is_same<T, typename unrangetype<T>::type>::value)
        {
            buffer_map<T> map(*this, begin, end);
            std::fill(map.begin(), map.end(), value);
            ret.did_finish = true;
        }
        else
        {
            rundata* rd =
                if_call(interface::goopax_41E_fill_buffer, this->get_buf(), &value,
                        sizeof(T), begin * sizeof(T), end * sizeof(T));
            ret.set(*rd);
        }
        return ret;
    }
    goopax_future<void> fill(T value)
    {
        /// Fill buffer with value
        /// \param value value to fill
        return fill(value, 0, this->size());
    }

    template<typename = void>
    T min(const size_type begin, const size_type end) const
    {
        /// Finds minimum value
        return utils::min(*this, begin, end);
    }
    template<typename = void>
    T min() const
    {
        /// Finds minimum value
        return min(0, this->size());
    }

    template<typename = void>
    T max(const size_type begin, const size_type end) const
    {
        /// Finds maximum value
        return utils::max(*this, begin, end);
    }
    template<typename = void>
    T max() const
    {
        /// Finds maximum value
        return max(0, this->size());
    }

    template<typename = void>
    T sum(const size_type begin, const size_type end) const
    {
        /// Calculates the sum of all values in the given range
        return utils::sum(*this, begin, end);
    }
    template<typename = void>
    T sum() const
    {
        /// Calculates the sum of all values
        return sum(0, this->size());
    }

    buffer& operator=(initializer_list<T> il)
    {
        /// assign data via initialization list
        return (*this = vector<T>(il));
    }

    vector<value_type> to_vector(size_type begin, size_type end) const
    {
        /// returns a std::vector that holds a copy of the data
        vector<value_type> ret(end - begin);
        copy_to_host(ret.data(), begin, end);
        return ret;
    }

    vector<value_type> to_vector() const
    {
        return to_vector(0, this->size());
    }

    template<typename... ARGS>
    void assign(ARGS&&... args)
    {
        *this = buffer(args...);
    }

    template<class V, class ENABLE = typename enable_if<std::is_base_of<
                          std::random_access_iterator_tag,
                          typename std::iterator_traits<typename V::iterator>::
                              iterator_category>::value>::type>
    [[deprecated("copy construction from lvalue vector will be "
                 "removed.")]] buffer(goopax_device device, const V& v)
        : buffer(device, V(v))
    {}

    template<class V, class ENABLE = typename enable_if<std::is_base_of<
                          std::random_access_iterator_tag,
                          typename std::iterator_traits<typename V::iterator>::
                              iterator_category>::value>::type>
    buffer(goopax_device device, V&& v)
        : buffer(device, v.size())
    {
        /// initialize from std::vector
        *this = move(v);
    }

    template<class IT, typename = typename std::iterator_traits<IT>::value_type>
    buffer(goopax_device device, const IT& a, const IT& b)
        : buffer(device,
                 vector<typename std::iterator_traits<IT>::value_type>(a, b))
    {
        /// initialize from begin/end pointer
    }

    buffer(goopax_device device, initializer_list<value_type> il)
        : buffer(device, vector<value_type>(il))
    {
        /// initialize from initialization list
    }

    buffer(goopax_device device, size_t size, T* host_ptr = nullptr,
           BUFFER_FLAGS flags = BUFFER_READ_WRITE)
        : base(device, size * (gettype<T>::size / gettype<base_value_type>::size),
               flags, nullptr, host_ptr, &mapframe<T>::get_static_mf(),
               &mapframe<typename unrangetype<T>::type>::get_static_mf())
    {
        /// initialize with size
        /// \param device device on which to allocate memory
        /// \param host_ptr pointer to host memory. If supplied, the buffer will be
        /// linked to the given address in host memory. host_ptr must point to a
        /// memory region that is at least size*sizeof(T) bytes long. When using the Metal
        /// backend, host_ptr must be aligned and large enough to fit whole memory pages (see `getpagesize()`).
        /// The data may
        /// or may not be buffered during a kernel call, depending on the device
        /// driver. After the kernel has finished, the data is synchronized and can
        /// be accessed by the host. \param flags access flags
    }

    private:
    buffer(texbuf& buf, INITTEX_REINTERPRET)
        : base(buf, INITTEX_REINTERPRET())
    {}

    public:
    /// move semantics
    buffer(buffer&&) = default;
    /// move semantics
    buffer& operator=(buffer&&) = default;

    static inline buffer create_from_gl(goopax_device device, unsigned int GLres,
                                        BUFFER_FLAGS flags = BUFFER_READ_WRITE);
#if GOOPAX_HAVE_METAL && defined(__OBJC__)
    static inline buffer create_from_metal(goopax_device device,
                                           id<MTLBuffer> tex)
    {
        return buffer(device, tex);
    }
#endif

#if GOOPAX_HAVE_OPENCL
    static inline buffer create_from_cl(goopax_device device, _cl_mem* mem)
    {
        ///\ingroup OpenCL
        /// create buffer from OpenCL memory buffer
        /// \param device GOOPAX device
        /// \param mem OpenCL memory buffer
        return buffer(device, mem);
    }
#endif
#if GOOPAX_HAVE_CUDA
    static inline buffer create_from_cuda(goopax_device device, void* mem,
                                          size_t size)
    {
        /// \ingroup CUDA
        /// create buffer from CUDA buffer
        /// \param device GOOPAX device
        /// \param pointer to CUDA device memory
        /// \param size size in number of elements of type T
        return buffer(device, mem, size);
    }
#endif

    /// default constructor. The buffer cannot be used until replaced by a buffer
    /// that has been properly initialized.
    buffer() = default;

    private:
    buffer(goopax_device device, size_type size, BUFFER_FLAGS flags,
           unsigned int* GLres, INITTEX_FROMGL)
        : base(device, size * (gettype<T>::size / gettype<base_value_type>::size),
               flags, GLres, nullptr, &mapframe<T>::get_static_mf(),
               &mapframe<typename unrangetype<T>::type>::get_static_mf())
    {}
#if GOOPAX_HAVE_OPENCL
    template<typename = void>
    buffer(goopax_device device, _cl_mem* mem)
        : base(device, mem)
    {}
#endif
#if GOOPAX_HAVE_METAL && defined(__OBJC__)
    template<typename = void>
    buffer(goopax_device device, void* mem)
        : base(device, mem)
    {}
#endif
#if GOOPAX_HAVE_CUDA
    template<typename = void>
    buffer(goopax_device device, void* mem, size_t size)
        : base(device, mem, size)
    {}
#endif
};

/// \ingroup memory
/// This class is used for local and private memory.
template<typename T, memory::scope SCOPE>
class local_scope_mem
    : private gpu_struct::resource_type_switch<T, unsigned int,
                                               SCOPE>::lds_type
{
    private:
    using base_t = typename gpu_struct::resource_type_switch<T, unsigned int,
                                                             SCOPE>::lds_type;
    unsigned int size_i;

    public:
    using typename base_t::size_type;
    using typename base_t::ssize_type;
    using cpu_size_type = typename gettype<size_type>::cpu;
    using typename base_t::const_reference;
    using typename base_t::reference;
    using typename base_t::value_type;

    using iterator       = gpu_type<T*, SCOPE>;
    using const_iterator = gpu_type<const T*, SCOPE>;

    using base_t::       operator[];
    inline cpu_size_type size() const
    {
        /// size of local/private memory
        return size_i;
    }
    using base_t::barrier;

    iterator begin()
    {
        /// returns begin() pointer
        iterator ret;
        check_errno(
            interface::goopax_4E_get_ptr(this->base.impl.get(), ret.impl.get()));
        return ret;
    }
    const_iterator begin() const
    {
        /// returns begin() pointer
        const_iterator ret;
        check_errno(
            interface::goopax_4E_get_ptr(this->base.impl.get(), ret.impl.get()));
        return ret;
    }
    iterator end()
    {
        /// returns end() iterator
        return begin() + this->size();
    }
    const_iterator end() const
    {
        /// returns end() iterator
        return begin() + this->size();
    }

    const_iterator cbegin() const
    {
        return begin();
    }
    const_iterator cend() const
    {
        return end();
    }

    private:
    template<typename R>
    void copy_impl(const R& r, size_type count, typename R::size_type start_dest,
                   size_type start_src)
    {
        if (SCOPE == memory::threadgroup)
        {
            gpu_for_local(0, count, [&](size_type k) {
                (*this)[start_dest + k] = r[start_src + k];
            });
        }
        else if (SCOPE == memory::thread)
        {
            gpu_for(0, count,
                    [&](size_type k) { (*this)[start_dest + k] = r[start_src + k]; });
        }
    }

    public:
    template<typename RESOURCE_SIZE_TYPE, bool is_const>
    void
    copy(const resource<T, RESOURCE_SIZE_TYPE, is_const>& r, size_type count,
         typename resource<T, RESOURCE_SIZE_TYPE, is_const>::size_type start_dest,
         size_type                                                     start_src)
    {
        /// copy data from resource. For local memory, the copy operation is done in
        /// parallel in the workgroup. For private memory, each thread performs its
        /// own copy operation. \param r resource to copy data from \param count
        /// size number of values to copy \param start_dest offset in local/private
        /// memory \param, start_src offset in the source resource
        copy_impl(r, count, start_dest, start_src);
    }

    template<typename RESOURCE_SIZE_TYPE, bool is_const>
    void copy(const resource<T, RESOURCE_SIZE_TYPE, is_const>& r)
    {
        /// copy data from resource. For local memory, the copy operation is done in
        /// parallel in the workgroup. For private memory, each thread performs its
        /// own copy operation. \param r resource to copy data from
        copy(r, this->size(), 0, 0);
    }

    template<memory::scope BSCOPE>
    void copy(const local_scope_mem<T, BSCOPE>& b, size_t count,
              typename local_scope_mem<T, BSCOPE>::size_type start_dest,
              size_type                                      start_src)
    {
        /// copy data from another local/private memory. For local memory, the copy
        /// operation is done in parallel in the workgroup. For private memory, each
        /// thread performs its own copy operation. \param b local/private memory to
        /// copy data from \param count number of elements to copy \param start_dest
        /// offset in destination memory \param start_src offset in source memory
        copy_impl(b, count, start_dest, start_src);
    }
    template<memory::scope BSCOPE>
    void copy(const local_scope_mem<T, BSCOPE>& b)
    {
        /// copy data from another local/private memory. For local memory, the copy
        /// operation is done in parallel in the workgroup. For private memory, each
        /// thread performs its own copy operation. \param b local/private memory to
        /// copy data from
        copy_impl(b, this->size(), 0, 0);
    }

    void fill(value_type value)
    {
        /// fill local/private memory. For local memory, the fill operation is done
        /// in parallel over the workgroup. \param value value to fill memory with
        gpu_for(((SCOPE == memory::threadgroup) ? local_id() : 0), this->size(),
                ((SCOPE == memory::threadgroup) ? local_size() : 1u),
                [&](size_type k) { (*this)[k] = value; });
    }

    /// move semantics
    local_scope_mem(local_scope_mem&&) = default;
    /// move semantics
    local_scope_mem& operator=(local_scope_mem&&) = default;

    local_scope_mem(typename gettype<size_type>::cpu size)
        : base_t(size * (gettype<T>::size / gettype<typename base_t::BASE_t::value_type>::size))
        , size_i(size)
    {
        /// constructor.
        /// \param size size of the local/private memory
    }
};
/// \ingroup memory
/// local memory. This memory can be accessed by all threads in the workgroup.
template<typename T>
using local_mem = local_scope_mem<T, memory::threadgroup>;
/// \ingroup memory
/// private memory. Each thread has its own private memory. No other thread can
/// access it.
template<typename T>
using private_mem = local_scope_mem<T, memory::thread>;

}

using impl::buffer;
using impl::buffer_map;
using impl::const_buffer_map;
using impl::const_resource;
using impl::local_mem;
using impl::private_mem;
using impl::resource;

}
