// -*- C++ -*-

#pragma once

#include "lib/range.h"
#include "lib/vec.h"

#include "env.h"
#include "future.h"
#include "gpu_defs.h"
#include "il_writer_utils.h"
#include "image.h"
#include "loops.h"
#include "reinterpret.h"

/**
 \addtogroup kernel
GPU kernels are inherited from kernel. The entry point to the GPU kernel is provided by the function 'program'. However, GPU kernels can call other functions, so they are not limited to the body of function 'program'.

 Example:
 \code{.cpp}
 struct my_kernel :
    kernel<my_kernel>
 {
    void program(resource<int>& A)
    {
        A[global_id()] = 5;   // This is executed on the GPU
    }
 };
 \endcode
\{
*/

namespace goopax {
namespace impl {

static constexpr unsigned int calc_lcd(const unsigned int a, const unsigned int b, const unsigned int loop = 2)
{
    return ((a % loop == 0) && (b % loop == 0)) ? loop * calc_lcd(a / loop, b / loop, loop)
                                                : ((loop >= a || loop >= b) ? a * b
                                                                            : calc_lcd(a, b, loop + 1));
}


template<typename A, typename B>
struct tuple_append;

template<typename B, typename... A>
struct tuple_append<tuple<A...>, B>
{
    using type = tuple<A..., B>;
};

using std::get;
using std::tuple_size;


template<class T>
struct unbool
{
    using type = T;
};
template<>
struct unbool<gpu_bool>
{
    using type = gpu_int;
};
template<>
struct unbool<bool>
{
    using type = int;
};


template<typename T>
struct debug_wrapper
{
    T data;

    debug_wrapper(T d)
        : data(d)
    {
    }

    debug_wrapper(typename goopax_struct_changetype<T, debug::types::Tuint>::type d)
        : data(reinterpret<typename goopax_struct_changetype<T, unsigned int>::type>(d))
    {
    }
};


namespace interface {
extern "C" {

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_fk_force_compile(kernel_impl*, goopax_device_impl*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_fk_arun(rundata**, goopax_device_impl*, kernel_impl*, texbuf* const*, unsigned int, const paramtype_t*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_fk_read_param(gpu_writer_base_impl*, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_fk_write_gather(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
#if !GOOPAX_DISTRIB
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_fk_get_disassembly(const char**, goopax_device_impl*, kernel_impl*) noexcept;
#endif
GOOPAX_DLL_EXPORT(4.0.0) unsigned int goopax_4_fk_get_num_writeloops(const kernel_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) unsigned int goopax_4_fk_get_current_writeloop(const kernel_impl*) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_fk_require_writeloops(kernel_impl*, unsigned int) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_fk_force_set(kernel_impl*, goopax_device_impl*, int, unsigned int) noexcept;
GOOPAX_DLL_EXPORT(4.1.1) errno_type goopax_4E_fk_get(kernel_impl*, goopax_device_impl*, int, void*) noexcept;

GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_fk_new(kernel_impl**, full_kernel*, const char*, const uint8_t*, void (*)(full_kernel*, unsigned int*), const char*, unsigned int, unsigned int) noexcept;
[[deprecated]] GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_fk_set_this(kernel_impl*, full_kernel*) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) void goopax_4_fk_add_this(kernel_impl*, full_kernel*) noexcept;
GOOPAX_DLL_EXPORT(4.1.0) void goopax_4_fk_remove_this(kernel_impl*, full_kernel*) noexcept;

GOOPAX_DLL_EXPORT(4.0.2) void goopax_4_fk_user_exception() noexcept;
GOOPAX_DLL_EXPORT(4.1.0) void goopax_4_fk_add_collect(unsigned int, void (&collect0)(const paramtype_t*, unsigned int, void*));
}
}

struct pas_data
{

    template<class V>
    static constexpr unsigned int paramsize()
    {
        return (calc_lcd(get_size<typename unbool<V>::type>::value, get_size<paramtype_t>::value) / get_size<paramtype_t>::value);
    }
    template<class V>
    static V read(const unsigned int parampos)
    {
        static const size_t                                                                                lcd = calc_lcd(gettype<typename unbool<V>::type>::size, gettype<paramtype_t>::size);
        array<typename gettype<V>::template change<paramtype_t>::type, (lcd / gettype<paramtype_t>::size)> data;
        for (unsigned int k = 0; k < data.size(); ++k)
        {
            interface::goopax_4_fk_read_param(data[k].impl.get(), parampos + k);
        }
        return read<V>(data, 0);
    }
    template<class V, typename POS, class PD>
    static V read(const PD& paramres, const POS parampos)
    {
        static const size_t                                                                                lcd = calc_lcd(gettype<typename unbool<V>::type>::size, gettype<paramtype_t>::size);
        array<typename gettype<V>::template change<paramtype_t>::type, (lcd / gettype<paramtype_t>::size)> data;
        for (unsigned int k = 0; k < data.size(); ++k)
        {
            data[k] = paramres[parampos + k];
        }
        auto rettab = reinterpret<array<typename unbool<V>::type, (lcd / gettype<typename unbool<V>::type>::size)>>(data);
        return static_cast<V>(rettab[0]);
    }

    template<class T>
    static void write_gather(const gpu_uint parampos, const T& value)
    {
        static const size_t                                                                                lcd = calc_lcd(gettype<typename unbool<T>::type>::size, gettype<paramtype_t>::size);
        array<typename gettype<T>::template change<paramtype_t>::type, (lcd / gettype<paramtype_t>::size)> data;
        write(data, 0, value);
        for (unsigned int k = 0; k < data.size(); ++k)
        {
            gpu_uint pos = parampos + k;
            interface::goopax_4_fk_write_gather(pos.impl.get(), data[k].impl.get());
        }
    }

    template<class T, typename POS, class PD>
    static void write(PD& paramdata, const POS parampos, const T& value)
    {
        using pt_t                                                                           = typename gettype<T>::template change<paramtype_t>::type;
        static const size_t                                                              lcd = calc_lcd(gettype<typename unbool<T>::type>::size, gettype<paramtype_t>::size);
        array<typename unbool<T>::type, (lcd / gettype<typename unbool<T>::type>::size)> valuetab;
        valuetab.fill(value);
        valuetab[0] = value;

        auto data = reinterpret<array<pt_t, (lcd / gettype<pt_t>::size)>>(valuetab);
        for (unsigned int k = 0; k < data.size(); ++k)
        {
            paramdata[parampos + k] = data[k];
        }
    }
};


template<typename = void>
void fk_call_impl(full_kernel*, unsigned int*);


template<typename OP>
typename OP::value_type collectfunc(const paramtype_t* p, size_t threads)
{
    auto ret = pas_data::read<typename OP::value_type>(p, 0);
    for (unsigned int k = 1; k < threads; ++k)
    {
        ret = OP::op(ret, pas_data::read<typename OP::value_type>(p, k * pas_data::paramsize<typename OP::value_type>()));
    }
    return ret;
}

/// \ingroup kernel
/// kernel base
struct full_kernel
{
    kernel_impl* impl_p = nullptr;

    /// \name overriding default thread numbers
    ///\{

    /// set the local size for this kernel for the specified device.
    inline void force_local_size(goopax_device device, int local_size) { interface::goopax_4_fk_force_set(impl_p, device.get_impl(), 0, local_size); }
    /// set the global size for this kernel for the specified device.
    inline void force_global_size(goopax_device device, int global_size) { interface::goopax_4_fk_force_set(impl_p, device.get_impl(), 1, global_size); }
    /// \}

    inline string name() const { return pretty_typename(typeid(*this)); }
    /// force compilation of kernel. Only useful for performance measurements and other occasions.
    inline void force_compile(goopax_device device) { check_errno(interface::goopax_4E_fk_force_compile(impl_p, device.get_impl())); }

#if !GOOPAX_DISTRIB
    inline string get_disassembly(goopax_device device) const
    {
        return if_call(interface::goopax_4E_fk_get_disassembly, device.get_impl(), impl_p);
    }
#endif

    inline unsigned int num_writeloops() const
    {
        return interface::goopax_4_fk_get_num_writeloops(impl_p);
    }
    inline unsigned int writeloop() const { return interface::goopax_4_fk_get_current_writeloop(impl_p); }
    inline void         require_writeloops(unsigned int num) const { interface::goopax_4_fk_require_writeloops(impl_p, num); }
    virtual void        implcall(unsigned int* p) = 0;

    full_kernel() = default;

    protected:
    inline void inc()
    {
        if (impl_p != nullptr)
        {
            interface::goopax_4_fk_add_this(impl_p, this);
        }
    }
    inline void dec()
    {
        if (impl_p != nullptr)
        {
            interface::goopax_4_fk_remove_this(impl_p, this);
        }
    }

    public:
    /// move semantics
    inline full_kernel(const full_kernel& b) noexcept = default;
    /// move semantics
    inline full_kernel& operator=(const full_kernel& b) noexcept = default;

    virtual ~full_kernel() = default;
};


template<typename>
void fk_call_impl(full_kernel* K, unsigned int* p)
{
    try
    {
        K->implcall(p);
    }
    catch (...)
    {
        exception_holder<>::current_exception() = std::current_exception();
        interface::goopax_4_fk_user_exception();
    }
}

struct gather_basecheck
{
};

template<class T>
struct gather_base : public gather_basecheck
    , gettype<T>::gpu
{
    using gpu_value_type = typename gettype<T>::gpu;
    using value_type     = T;

    gather_base()
    {}

    gather_base(const gpu_value_type& val)
        : gpu_value_type(val)
    {}

    gather_base(const gather_base&) = default;

    //virtual ~gather_base() = default;
};

template<typename T>
struct is_gather_type : is_base_of<gather_basecheck, T>
{};


namespace ns_param_kernel {
template<class T>
struct is_resource_type : public false_type
{};

template<class T, typename size_type, bool is_const>
struct is_resource_type<resource<T, size_type, is_const>> : public true_type
{};

template<unsigned int DIM, typename T, bool normalized>
struct is_resource_type<image_resource<DIM, T, normalized>> : public true_type
{};
template<unsigned int DIM, typename T, bool normalized>
struct is_resource_type<image_array_resource<DIM, T, normalized>> : public true_type
{};

typedef unsigned int pas_base_type;

template<class R, class gpu_T, class REST>
struct pas_split_common
{
    static_assert(is_same<gpu_T, typename gettype<gpu_T>::gpu>::value, "Bad kernel parameter type. Perhaps you used an intrinsic type where you should have a gpu type in the program(...) function of your kernel?");
    typedef typename unrangetype<typename gettype<gpu_T>::cpu>::type T;
    static constexpr unsigned int                                    param_size() { return pas_data::paramsize<gpu_T>() + pas<R, REST>::param_size(); }
    static constexpr unsigned int                                    gather_size() { return pas<R, REST>::gather_size(); }
    static constexpr unsigned int                                    gather_num() { return pas<R, REST>::gather_num(); }
    static const unsigned int                                        num_tex_args = pas<R, REST>::num_tex_args;
    template<class K, class... P>
    static void impl(K& kernel, P&... p)
    {
        const auto data = pas_data::read<gpu_T>(pas<R, REST>::param_size());
        pas<R, REST>::impl(kernel, p..., data);
    }
    template<class K, class V, typename PARAM, class RP, class... P>
    static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, size_t vecpos, const T& b, P... p)
    {
        pas_data::write(param, pas<R, REST>::param_size(), b);
        pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos, p...);
    }
    template<class K, class V, typename PARAM, class RP, class... P>
    static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, size_t vecpos, const debug_wrapper<T>& b, P... p)
    {
        pas_data::write(param, pas<R, REST>::param_size(), b.data); // FIXME: TODO: enable passing of uninitialized checknan types.
        pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos, p...);
    }
    template<class RP, typename U, class... P>
    static inline void setcollect(rundata& rd, RP rp, const U&, P... p)
    {
        pas<R, REST>::setcollect(rd, rp, p...);
    }
};

template<class R, class MEM, class REST>
struct pas_split<R, MEM, REST, false, true, false>
{
    static_assert(sizeof(MEM) == 0, "Please pass resources as references or const references to the GPU kernel!\ne.g.: program(resource<int>& a);");
};


template<typename BUFREF, typename R, typename MEM, typename REST>
struct pas_common_resource
{
    static constexpr unsigned int param_size() { return pas<R, REST>::param_size(); }
    static constexpr unsigned int gather_size() { return pas<R, REST>::gather_size(); }
    static constexpr unsigned int gather_num() { return pas<R, REST>::gather_num(); }
    static const unsigned int     num_tex_args = pas<R, REST>::num_tex_args + 1;
    template<class K, class... P>
    static void impl(K& kernel, P&... p)
    {
        MEM res;
        pas<R, REST>::impl(kernel, p..., res);
    }
    template<class K, class V, typename PARAM, class RP, class... P>
    static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, size_t vecpos, BUFREF b, P... p)
    {
        vec[vecpos] = b.get_buf();
        pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos + 1, p...);
    }
    template<class RP, class... P>
    static inline void setcollect(rundata& rd, RP rp, BUFREF, P... p)
    {
        pas<R, REST>::setcollect(rd, rp, p...);
    }
};


template<class R, typename MEM1, typename MEM2, class REST>
struct pas_split<R, resource<MEM1, MEM2, false>&, REST, false, true, false> : public pas_common_resource<buffer<MEM1, MEM2>&, R, resource<MEM1, MEM2, false>, REST>
{};

template<class R, typename MEM1, typename MEM2, class REST>
struct pas_split<R, resource<MEM1, MEM2, true>&, REST, false, true, false> : public pas_common_resource<const buffer<MEM1, MEM2>&, R, resource<MEM1, MEM2, true>, REST>
{};

template<class R, typename MEM1, typename MEM2, class REST, bool is_const>
struct pas_split<R, const resource<MEM1, MEM2, is_const>&, REST, false, true, false> : public pas_common_resource<const buffer<MEM1, MEM2>&, R, resource<MEM1, MEM2, is_const>, REST>
{};

template<typename BUFREF, class R, typename IMAGE, class REST>
struct pas_common_image
{
    static constexpr unsigned int param_size() { return pas<R, REST>::param_size(); }
    static constexpr unsigned int gather_size() { return pas<R, REST>::gather_size(); }
    static constexpr unsigned int gather_num() { return pas<R, REST>::gather_num(); }
    static const unsigned int     num_tex_args = pas<R, REST>::num_tex_args + 1;
    template<class K, class... P>
    static void impl(K& kernel, P&... p)
    {
        IMAGE res;
        pas<R, REST>::impl(kernel, p..., res);
    }
    template<class K, class V, typename PARAM, class RP, class... P>
    static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, size_t vecpos, BUFREF b, P... p)
    {
        vec[vecpos] = b.get_buf();
        pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos + 1, p...);
    }
    template<class RP, class... P>
    static inline void setcollect(rundata& rd, RP rp, BUFREF, P... p)
    {
        pas<R, REST>::setcollect(rd, rp, p...);
    }
};

template<class R, unsigned int DIM, typename T, bool normalized, class REST>
struct pas_split<R, image_resource<DIM, T, normalized>&, REST, false, true, false> : public pas_common_image<image_buffer<DIM, T, normalized>&, R, image_resource<DIM, T, normalized>, REST>
{};

template<class R, unsigned int DIM, typename T, bool normalized, class REST>
struct pas_split<R, const image_resource<DIM, T, normalized>&, REST, false, true, false> : public pas_common_image<const image_buffer<DIM, T, normalized>&, R, image_resource<DIM, T, normalized>, REST>
{};

template<class R, unsigned int DIM, typename T, bool normalized, class REST>
struct pas_split<R, image_array_resource<DIM, T, normalized>&, REST, false, true, false> : public pas_common_image<image_array_buffer<DIM, T, normalized>&, R, image_array_resource<DIM, T, normalized>, REST>
{};

template<class R, unsigned int DIM, typename T, bool normalized, class REST>
struct pas_split<R, const image_array_resource<DIM, T, normalized>&, REST, false, true, false> : public pas_common_image<const image_array_buffer<DIM, T, normalized>&, R, image_array_resource<DIM, T, normalized>, REST>
{};


template<class R, class gpu_T, class REST>
struct pas_split<R, gpu_T, REST, true, false, false> : public pas_split_common<R, gpu_T, REST>
{};

template<class R, class gpu_T, class REST>
struct pas_split<R, gpu_T, REST, false, false, false> : public pas_split_common<R, gpu_T, REST>
{};


template<class R, class G, class REST>
struct pas_split<R, G&, REST, false, false, true>
{
    using gather_type = typename G::value_type;
    static constexpr unsigned int param_size() { return pas<R, REST>::param_size(); }
    static constexpr unsigned int gather_size() { return pas<R, REST>::gather_size() + pas_data::paramsize<gather_type>(); }
    static constexpr unsigned int gather_num() { return pas<R, REST>::gather_num() + 1; }
    static const unsigned int     num_tex_args = pas<R, REST>::num_tex_args;

    template<class K, class... P>
    static void impl(K& kernel, P&... p)
    {
        G val;
        pas<R, REST>::impl(kernel, p..., val);
        typename gettype<gather_type>::gpu tot = work_group_reduce<G>((typename gettype<gather_type>::gpu)val, local_size());
        gpu_if(local_id() == 0)
            pas_data::write_gather(pas<R, REST>::gather_size() * num_groups() + group_id() * pas_data::paramsize<gather_type>(), tot);
    }
    template<class K, class V, typename PARAM, class RP, class... P>
    static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, size_t vecpos, goopax_future<gather_type>&, P... p)
    {
        pas<R, REST>::prepare(device_p, kernel, param, rp, vec, vecpos, p...);
    }
    template<class RP, class... P>
    static inline void setcollect(rundata& rd, RP rp, goopax_future<gather_type>& b, P... p)
    {
        pas<R, REST>::setcollect(rd, rp, p...);
        b = goopax_future<gather_type>();
        b.set(rd, pas<R, REST>::gather_size(), collectfunc<G>);
    }
};


template<class T, class R, class... A>
struct pas<R, tuple<T, A...>>
{
    using raw_T   = typename remove_const<typename remove_reference<T>::type>::type;
    using split_t = pas_split<R, T, tuple<A...>, is_gpu_type<raw_T>::value, is_resource_type<raw_T>::value, is_gather_type<raw_T>::value>;
    static constexpr unsigned int param_size() { return split_t::param_size(); }
    static constexpr unsigned int gather_size() { return split_t::gather_size(); }
    static constexpr unsigned int gather_num() { return split_t::gather_num(); }
    static const unsigned int     num_tex_args = split_t::num_tex_args;
    template<class... P>
    static void impl(P&... p)
    {
        split_t::impl(p...);
    }
    template<class K, class V, typename PARAM, class E, class RP, class... P>
    static inline void prepare(goopax_device_impl* device_p, K& kernel, PARAM& param, RP rp, V& vec, size_t vecpos, E e, P... p)
    {
        split_t::prepare(device_p, kernel, param, rp, vec, vecpos, *e, p...);
    }
    template<class E, class RP, class... P>
    static inline void setcollect(rundata& rd, RP rp, E e, P... p)
    {
        split_t::setcollect(rd, rp, *e, p...);
    }
};


template<>
struct pas<void, tuple<>>
{
    using return_type = void;
    static constexpr unsigned int param_size() { return 0; }
    static constexpr unsigned int gather_size() { return 0; }
    static constexpr unsigned int gather_num() { return 1; }
    static constexpr unsigned int num_tex_args = 0;

    template<class K, class... P>
    static void impl(K& kernel, P&... p)
    {
        kernel.program(p...);
    }
    template<class K, class V, typename PARAM>
    static inline void prepare(goopax_device_impl*, K&, PARAM&, void*, V&, size_t)
    {
    }
    static inline void setcollect(rundata& rd, goopax_future<void>* rp)
    {
        rp->set(rd);
    }
};


template<class R>
struct pas<R, tuple<>>
{
    using return_type = typename R::value_type;
    static constexpr unsigned int param_size() { return 0; }
    static constexpr unsigned int gather_size() { return pas_data::paramsize<return_type>(); }
    static constexpr unsigned int gather_num() { return 1; }
    static constexpr unsigned int num_tex_args = 0;

    template<class K, class... P>
    static void impl(K& kernel, P&... p)
    {
        auto                               ret = kernel.program(p...);
        typename gettype<return_type>::gpu tot = work_group_reduce<R>((typename gettype<return_type>::gpu)ret, local_size());
        gpu_if(local_id() == 0)
            pas_data::write_gather(group_id() * pas_data::paramsize<return_type>(), tot);
    }
    template<class K, class V, typename PARAM, class RP>
    static inline void prepare(goopax_device_impl*, K&, PARAM&, RP, V&, size_t)
    {
    }
    template<class RP>
    static inline void setcollect(rundata& rd, RP rp)
    {
        rp->set(rd, 0, collectfunc<R>);
    }
};


template<class G>
struct return_type_wrapper
{
    using type = typename G::value_type;
};
template<>
struct return_type_wrapper<void>
{
    using type = void;
};


template<class FUNC>
struct funcinfo;

template<class K, class R, class... A>
struct funcinfo<R (K::*)(A...)>
{
    using params      = tuple<A...>;
    using return_type = R;
    static constexpr unsigned int param_size() { return pas<R, tuple<A...>>::param_size(); }
    static constexpr unsigned int gather_size() { return pas<R, tuple<A...>>::gather_size(); }

    static void impl(K& kernel)
    {
        pas<R, tuple<A...>>::impl(kernel);
    }
};


template<int MODE, typename T, bool is_gather, bool is_struct>
struct change_type_single;

template<int MODE, typename T, typename SIZE>
struct change_type_single<MODE, resource<T, SIZE, false>, false, false>
{
    using type                           = buffer<T, SIZE>;
    static constexpr bool can_get_device = true;
};

template<int MODE, typename T, typename SIZE>
struct change_type_single<MODE, resource<T, SIZE, true>, false, false>
{
    using type                           = const buffer<T, SIZE>;
    static constexpr bool can_get_device = true;
};

template<int MODE, unsigned int DIM, typename T, bool is_normalized>
struct change_type_single<MODE, image_resource<DIM, T, is_normalized>, false, false>
{
    using type                           = image_buffer<DIM, T, is_normalized>;
    static constexpr bool can_get_device = true;
};

template<int MODE, unsigned int DIM, typename T, bool is_normalized>
struct change_type_single<MODE, image_array_resource<DIM, T, is_normalized>, false, false>
{
    using type                           = image_array_buffer<DIM, T, is_normalized>;
    static constexpr bool can_get_device = true;
};

template<int MODE, typename T>
struct change_type_single<MODE, T&, false, false>
{
    using type                           = typename change_type_single<MODE, T, false, false>::type&;
    static constexpr bool can_get_device = change_type_single<MODE, T, false, false>::can_get_device;
};
template<int MODE, typename T>
struct change_type_single<MODE, const T&, false, false>
{
    using type                           = const typename change_type_single<MODE, T, false, false>::type&;
    static constexpr bool can_get_device = change_type_single<MODE, T, false, false>::can_get_device;
};

template<int MODE, typename T>
struct change_type_single<MODE, T&, true, false>
{
    using type                           = goopax_future<typename T::value_type>&;
    static constexpr bool can_get_device = false;
};
template<int MODE, typename T>
struct change_type_single<MODE, gpu_type<T>, false, false>
{
    using type                           = const typename conditional<MODE == 1, debug_wrapper<T>, T>::type&;
    static constexpr bool can_get_device = false;
};

template<int MODE, typename T>
struct change_type_single<MODE, T, false, true>
{
    using raw_T                          = typename unrangetype<typename gettype<T>::cpu>::type;
    using type                           = const typename conditional<MODE == 1, debug_wrapper<raw_T>, raw_T>::type&;
    static constexpr bool can_get_device = false;
};


template<int MODE, typename...>
struct change_type;

template<int MODE, typename... HAVE, typename T, typename... B>
struct change_type<MODE, tuple<HAVE...>, tuple<T, B...>>
{
    using raw_T                                       = typename remove_const<typename remove_reference<T>::type>::type;
    using current                                     = change_type_single<MODE, T, is_gather_type<raw_T>::value, is_gpu_struct<raw_T>::value>;
    using rest                                        = change_type<MODE, tuple<HAVE..., typename current::type>, tuple<B...>>;
    static constexpr bool can_get_device_from_current = current::can_get_device;
    using type                                        = typename rest::type;
    static constexpr bool can_get_device              = can_get_device_from_current || rest::can_get_device;

    template<typename... C>
    static inline goopax_device_impl* get_device_split(true_type, typename current::type t, C&&...)
    {
        return t.get_device().get_impl();
    }
    template<typename... C>
    static inline goopax_device_impl* get_device_split(false_type, typename current::type, C&&... b)
    {
        return rest::get_device(b...);
    }
    template<typename... C>
    static inline goopax_device_impl* get_device(typename current::type t, C&&... b)
    {
        return get_device_split(integral_constant<bool, can_get_device_from_current>(), t, b...);
    }
};

template<int MODE, typename HAVE>
struct change_type<MODE, HAVE, tuple<>>
{
    using type                                       = HAVE;
    static constexpr bool             can_get_device = false;
    static inline goopax_device_impl* get_device()
    {
        abort();
    }
};


template<typename FUNC, typename P, typename D>
class kernelparams;
template<typename K, typename R, typename... A, typename... B, typename... D>
class kernelparams<R (K::*)(A...), tuple<B...>, tuple<D...>>
{
    public:
    using return_type = goopax_future<typename return_type_wrapper<R>::type>;

    static inline void arun(int, K& kernel, return_type* ret, B&&... p)
    {
        using ct                   = change_type<0, tuple<>, tuple<A...>>;
        goopax_device_impl* device = ct::get_device(p...);
        return arun(int(), kernel, device, ret, p...);
    }
    static inline void arun(int, K& kernel, goopax_device_impl* device, return_type* ret, B&&... p)
    {
        static constexpr unsigned int param_size   = pas<R, tuple<A...>>::param_size();
        static constexpr unsigned int gather_num   = pas<R, tuple<A...>>::gather_num();
        static constexpr unsigned int num_tex_args = pas<R, tuple<A...>>::num_tex_args;

        array<paramtype_t, param_size> pd;
        array<texbuf*, num_tex_args>   vec;
        pas<R, tuple<A...>>::prepare(device, kernel, pd, ret, vec, 0, &p...);

        rundata* rd = if_call(interface::goopax_4E_fk_arun, device, kernel.impl_p, vec.data(), (unsigned int)vec.size(), pd.data(), gather_num);

        pas<R, tuple<A...>>::setcollect(*rd, ret, &p...);
    }

    static inline void arun(float, K& kernel, return_type* ret, D&&... p)
    {
        using ct                   = change_type<1, tuple<>, tuple<A...>>;
        goopax_device_impl* device = ct::get_device(p...);
        return arun(int(), kernel, device, ret, p...);
    }

    static inline void arun(float, K& kernel, goopax_device_impl* device, return_type* ret, D&&... p)
    {
        static constexpr unsigned int param_size   = pas<R, tuple<A...>>::param_size();
        static constexpr unsigned int gather_num   = pas<R, tuple<A...>>::gather_num();
        static constexpr unsigned int num_tex_args = pas<R, tuple<A...>>::num_tex_args;

        array<paramtype_t, param_size> pd;
        array<texbuf*, num_tex_args>   vec;
        pas<R, tuple<A...>>::prepare(device, kernel, pd, ret, vec, 0, &p...);

        rundata* rd = if_call(interface::goopax_4E_fk_arun, device, kernel.impl_p, vec.data(), (unsigned int)vec.size(), pd.data(), gather_num);

        pas<R, tuple<A...>>::setcollect(*rd, ret, &p...);
    }
};
}

/// \ingroup kernel
/// \brief kernel class to derive gpu kernels from
/// \param K derived kernel class.
/// If The C++ language standard is < 14, any return type of the kernel must be specified with the RETURN_TYPE template argument.
/// All gpu kernels must be derived from this class, where the new class must be passed as template argument, like so:
/// struct my_kernel :
/// public kernel<my_kernel>
/// {
///    void program(...) {...}
/// };
template<class K>
class kernel : public full_kernel
{
    virtual void implcall(unsigned int* p) final override
    {
        using functype = decltype(&K::program);
        *p             = ns_param_kernel::funcinfo<functype>::param_size();
        ns_param_kernel::funcinfo<functype>::impl(static_cast<K&>(*this));
    }

    public:
    /// Operator (). This starts the gpu kernel on the device. The parameters must match the arguments specified in the function 'program()', exept that cpu types are passed, where gpu types are specified in the function 'program()'. The device may be specified as a first argument. This is optional except if the device cannot be deduced from the other arguments. All buffers must reside on the same device, which is the device on which the kernel is executed.
    template<class... A>
    inline auto operator()(goopax_device device, A&&... a)
    {
        using functype = decltype(&K::program);
        using ct       = ns_param_kernel::change_type<0, tuple<>, typename ns_param_kernel::funcinfo<functype>::params>;
        using ct1      = ns_param_kernel::change_type<1, tuple<>, typename ns_param_kernel::funcinfo<functype>::params>;
        using kp       = ns_param_kernel::kernelparams<decltype(&K::program), typename ct::type, typename ct1::type>;
        goopax_future<typename ns_param_kernel::return_type_wrapper<typename ns_param_kernel::funcinfo<functype>::return_type>::type> ret;
        kp::arun(int(), static_cast<K&>(*this), device.get_impl(), &ret, a...);
        return ret;
    }
    /// Operator (). This starts the gpu kernel on the device. The parameters must match the arguments specified in the function 'program()', exept that cpu types are passed, where gpu types are specified in the function 'program()'.
    /// The device is deduced from the buffers that are passed as arguments.
    template<class... A>
    inline auto operator()(A&&... a)
    {
        using functype = decltype(&K::program);
        using ct       = ns_param_kernel::change_type<0, tuple<>, typename ns_param_kernel::funcinfo<functype>::params>;
        using ct1      = ns_param_kernel::change_type<1, tuple<>, typename ns_param_kernel::funcinfo<functype>::params>;
        using kp       = ns_param_kernel::kernelparams<decltype(&K::program), typename ct::type, typename ct1::type>;
        static_assert(ct::can_get_device, "Cannot deduce device from input buffers. Please pass the goopax_device as first argument.");

        goopax_future<typename ns_param_kernel::return_type_wrapper<typename ns_param_kernel::funcinfo<functype>::return_type>::type> ret;
        kp::arun(int(), static_cast<K&>(*this), &ret, a...);

        return ret;
    }

    kernel()
    {
        const string name = pretty_typename(typeid(K));
        /// constructor
        this->impl_p = if_call(interface::goopax_4E_fk_new,
                               static_cast<full_kernel*>(this),
                               GOOPAX_CERT, sig,
                               fk_call_impl<>,
                               name.c_str(),
                               ns_param_kernel::funcinfo<decltype(&K::program)>::param_size(),
                               ns_param_kernel::funcinfo<decltype(&K::program)>::gather_size());
        full_kernel::inc();
    }
    kernel(const kernel& b)
        : full_kernel(b)
    {
        full_kernel::inc();
    }
    kernel& operator=(const kernel& b)
    {
        full_kernel::dec();
        full_kernel::operator=(b);
        full_kernel::inc();
        return *this;
    }
    ~kernel()
    {
        if (this->impl_p != nullptr)
        {
            wait_all_devices();
        }
        full_kernel::dec();
    }
};


template<typename FUNC, typename OP>
struct auto_kernel;

template<typename FUNC, typename ClassType, typename ReturnType, typename... Args>
struct auto_kernel<FUNC, ReturnType (ClassType::*)(Args...) const> : public kernel<auto_kernel<FUNC, ReturnType (ClassType::*)(Args...) const>>
{
    FUNC func;

    ReturnType program(Args... a)
    {
        return func(a...);
    }
    auto_kernel(FUNC f)
        : func(f)
    {}
};

template<typename FUNC, typename ClassType, typename... Args>
struct auto_kernel<FUNC, void (ClassType::*)(Args...) const> : public kernel<auto_kernel<FUNC, void (ClassType::*)(Args...) const>>
{
    FUNC func;

    void program(Args... a)
    {
        func(a...);
    }
    auto_kernel(FUNC f)
        : func(f)
    {}
};

template<class FUNC>
auto_kernel<FUNC, decltype(&FUNC::operator())> make_kernel(FUNC func)
{
    using OP = decltype(&FUNC::operator());

    return auto_kernel<FUNC, OP>(func);
}
}

using impl::full_kernel;
using impl::gather_base;
using impl::goopax_future;
using impl::kernel;
using impl::make_kernel;
}


/// \}
